@@ -1,5 +1,5 @@
 /* grep.c - main driver file for grep.
-   Copyright (C) 1992, 1997 Free Software Foundation, Inc.
+   Copyright (C) 1992, 1997, 1998 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -55,9 +55,11 @@ static struct option long_options[] =
   {"context", no_argument, NULL, 'C'},
   {"count", no_argument, NULL, 'c'},
   {"extended-regexp", no_argument, NULL, 'E'},
-  {"file", no_argument, NULL, 'f'},
+  {"file", required_argument, NULL, 'f'},
+  {"files-with-matches", no_argument, NULL, 'l'},
   {"files-without-match", no_argument, NULL, 'L'},
   {"fixed-regexp", no_argument, NULL, 'F'},
+  {"fixed-strings", no_argument, NULL, 'F'},
   {"help", no_argument, &show_help, 1},
   {"ignore-case", no_argument, NULL, 'i'},
   {"line-number", no_argument, NULL, 'n'},
@@ -65,7 +67,7 @@ static struct option long_options[] =
   {"no-filename", no_argument, NULL, 'h'},
   {"no-messages", no_argument, NULL, 's'},
   {"quiet", no_argument, NULL, 'q'},
-  {"regexp", no_argument, NULL, 'e'},
+  {"regexp", required_argument, NULL, 'e'},
   {"revert-match", no_argument, NULL, 'v'},
   {"silent", no_argument, NULL, 'q'},
 #if O_BINARY
@@ -73,6 +75,7 @@ static struct option long_options[] =
   {"unix-byte-offsets", no_argument, NULL, 'u'},
 #endif
   {"version", no_argument, NULL, 'V'},
+  {"with-filename", no_argument, NULL, 'H'},
   {"word-regexp", no_argument, NULL, 'w'},
   {0, 0, 0, 0}
 };
@@ -88,6 +91,7 @@ static char *prog;
 static char *filename;
 static int errseen;
 
+static int ck_atoi PARAMS((char const *, int *));
 static void usage PARAMS((int));
 static void error PARAMS((const char *, int));
 static int  setmatcher PARAMS((char *));
@@ -156,8 +160,26 @@ xrealloc(ptr, size)
   if (size && !result)
     fatal(_("memory exhausted"), 0);
   return result;
+
 }
 
+/* Convert STR to a positive integer, storing the result in *OUT.
+   If STR is not a valid integer, return -1 (otherwise 0). */
+static int
+ck_atoi (str, out)
+     char const *str;
+     int *out;
+{
+  char const *p;
+  for (p = str; *p; p++)
+    if (*p < '0' || *p > '9')
+      return -1;
+
+  *out = atoi (optarg);
+  return 0;
+}
+
+
 /* Hairy buffering mechanism for grep.  The intent is to keep
    all reads aligned on a page boundary and multiples of the
    page size. */
@@ -263,8 +285,14 @@ fillbuf(save)
 		   MAP_PRIVATE | MAP_FIXED, bufdesc, bufoffset);
       if (maddr == (caddr_t) -1)
 	{
+          /* This used to issue a warning, but on some hosts
+             (e.g. Solaris 2.5) mmap can fail merely because some
+             other process has an advisory erad lock on the file.
+             There's no point alarming the use about this misfeature */
+#if 0
 	  fprintf(stderr, _("%s: warning: %s: %s\n"), prog, filename,
 		  strerror(errno));
+#endif
 	  goto tryread;
 	}
 #if 0
@@ -572,48 +600,51 @@ int status;
   else
     {
       printf (_("Usage: %s [OPTION]... PATTERN [FILE] ...\n"), prog);
-      printf ("\n");
-      printf (_("Regexp selection and interpretation:\n"));
-      printf (_("  -E, --extended-regexp     PATTERN is an extended regular expression\n"));
-      printf (_("  -F, --fixed-strings       PATTERN is a fixed string separated by newlines\n"));
-      printf (_("  -G, --basic-regexp        PATTERN is a basic regular expression\n"));
-      printf (_("  -e, --regexp=PATTERN      use PATTERN as a regular expression\n"));
-      printf (_("  -f, --file=FILE         obtain PATTERN from FILE\n"));
-      printf (_("  -i, --ignore-case         ignore case distinctions\n"));
-      printf (_("  -w, --word-regexp         force PATTERN to match only whole words\n"));
-      printf (_("  -x, --line-regexp         force PATTERN to match only whole lines\n"));
-      printf ("\n");
-      printf (_("Miscellaneous:\n"));
-      printf (_("  -s, --no-messages         suppress error messages\n"));
-      printf (_("  -v, --revert-match        select non-matching lines\n"));
-      printf (_("  -V, --version             print version information and exit\n"));
-      printf (_("      --help                display this help and exit\n"));
-      printf ("\n");
-      printf (_("Output control:\n"));
-      printf (_("  -b, --byte-offset         print the byte offset with output lines\n"));
-      printf (_("  -n, --line-number         print line number with output lines\n"));
-      printf (_("  -H, --with-filename       print the filename for each match\n"));
-      printf (_("  -h, --no-filename         suppress the prefixing filename on ouput\n"));
-      printf (_("  -q, --quiet, --silent     suppress all normal output\n"));
-      printf (_("  -L, --files-without-match only print FILE names containing no match\n"));
-      printf (_("  -l, --files-with-matches  only print FILE names containing matches\n"));
-      printf (_("  -c, --count               only print a count of matching lines per FILE\n"));
-      printf ("\n");
-      printf (_("Context control:\n"));
-      printf (_("  -B, --before-context=NUM  print NUM lines of leading context\n"));
-      printf (_("  -A, --after-context=NUM   print NUM lines of trailing context\n"));
-      printf (_("  -NUM                      same as both -B NUM and -A NUM\n"));
-      printf (_("  -C, --context             same as -2\n"));
-#if O_BINARY
-      printf (_("  -U, --binary              do not strip CR characters at EOL\n"));
-      printf (_("  -u, --unix-byte-offsets   report offsets as if CRs were not there\n"));
-#endif
-      printf ("\n");
-      printf (_("There should be one and only one PATTERN, `-e PATTERN' or `-f FILE'.\n"));
-      printf (_("If call as `egrep', this implies -E and `fgrep' for -F.\n"));
-      printf (_("If no -[GEF], then -G is assumed.\n"));
-      printf ("\n");
-      printf (_("Report bugs to <bug-gnu-utils@prep.ai.mit.edu>.\n"));
+      printf (_("\
+Search for PATTERN in each FILE or standard input.\n\
+\n\
+Regexp selection and interpretation:\n\
+  -E, --extended-regexp     PATTERN is an extended regular expression\n\
+  -F, --fixed-regexp        PATTERN is a fixed string separated by newlines\n\
+  -G, --basic-regexp        PATTERN is a basic regular expression\n\
+  -e, --regexp=PATTERN      use PATTERN as a regular expression\n\
+  -f, --file=FILE           obtain PATTERN from FILE\n\
+  -i, --ignore-case         ignore case distinctions\n\
+  -w, --word-regexp         force PATTERN to match only whole words\n\
+  -x, --line-regexp         force PATTERN to match only whole lines\n"));
+      printf (_("\
+\n\
+Miscellaneous:\n\
+  -s, --no-messages         suppress error messages\n\
+  -v, --revert-match        select non-matching lines\n\
+  -V, --version             print version information and exit\n\
+      --help                display this help and exit\n"));
+      printf (_("\
+\n\
+Output control:\n\
+  -b, --byte-offset         print the byte offset with output lines\n\
+  -n, --line-number         print line number with output lines\n\
+  -H, --with-filename       print the filename for each match\n\
+  -h, --no-filename         suppress the prefixing filename on output\n\
+  -q, --quiet, --silent     suppress all normal output\n\
+  -L, --files-without-match only print FILE names containing no match\n\
+  -l, --files-with-matches  only print FILE names containing matches\n\
+  -c, --count               only print a count of matching lines per FILE\n"));
+      printf (_("\
+\n\
+Context control:\n\
+  -B, --before-context=NUM  print NUM lines of leading context\n\
+  -A, --after-context=NUM   print NUM lines of trailing context\n\
+  -NUM                      same as both -B NUM and -A NUM\n\
+  -C, --context             same as -2\n\
+  -U, --binary              do not strip CR characters at EOL (MSDOS)\n\
+  -u, --unix-byte-offsets   report offsets as if CRs were not there (MSDOS)\n\
+\n\
+If no -[GEF], then `egrep' assumes -E, `fgrep' -F, else -G.\n\
+With no FILE, or when FILE is -, read standard input. If less than\n\
+two FILEs given, assume -h. Exit with 0 if matches, with 1 if none.\n\
+Exit with 2 if syntax errors or system errors.\n"));
+      printf (_("\nReport bugs to <bug-gnu-utils@gnu.org>.\n"));
     }
   exit(status);
 }
@@ -671,7 +702,7 @@ main(argc, argv)
 {
   char *keys;
   size_t keycc, oldcc, keyalloc;
-  int keyfound, count_matches, no_filenames, list_files, suppress_errors;
+  int count_matches, no_filenames, list_files, suppress_errors;
   int with_filenames;
   int opt, cc, desc, count, status;
   FILE *fp;
@@ -709,7 +740,6 @@ main(argc, argv)
 
   keys = NULL;
   keycc = 0;
-  keyfound = 0;
   count_matches = 0;
   no_filenames = 0;
   with_filenames = 0;
@@ -749,20 +779,24 @@ main(argc, argv)
 	out_after = 10 * out_after + opt - '0';
 	break;
       case 'A':
-	out_after = atoi(optarg);
-	if (out_after < 0)
-	  usage(2);
+	if (optarg)
+	  {
+	    if (ck_atoi (optarg, &out_after))
+	      fatal (_("invalid context length argument"), 0);
+	  }
 	break;
       case 'B':
-	out_before = atoi(optarg);
-	if (out_before < 0)
-	  usage(2);
+	if (optarg)
+	  {
+	    if (ck_atoi (optarg, &out_before))
+	      fatal (_("invalid context length argument"), 0);
+	  }
 	break;
       case 'C':
 	out_before = out_after = 2;
 	break;
       case 'E':
-	if (matcher && strcmp(matcher, "egrep") != 0)
+	if (matcher && strcmp(matcher, "posix-egrep") != 0)
 	  fatal(_("you may specify only one of -E, -F, or -G"), 0);
 	matcher = "posix-egrep";
 	break;
@@ -804,36 +838,31 @@ main(argc, argv)
 	break;
       case 'e':
 	cc = strlen(optarg);
-	keys = xrealloc(keys, keycc + cc + 1 + keyfound);
-	if (keyfound)
-	  keys[keycc++] = '\n';
+	keys = xrealloc(keys, keycc + cc + 1);
 	strcpy(&keys[keycc], optarg);
 	keycc += cc;
-	keyfound = 1;
+	keys[keycc++] = '\n';
 	break;
       case 'f':
 	fp = strcmp(optarg, "-") != 0 ? fopen(optarg, "r") : stdin;
 	if (!fp)
 	  fatal(optarg, errno);
-	for (keyalloc = 1; keyalloc <= keycc + keyfound; keyalloc *= 2)
+	for (keyalloc = 1; keyalloc <= keycc + 1; keyalloc *= 2)
 	  ;
 	keys = xrealloc(keys, keyalloc);
 	oldcc = keycc;
-	if (keyfound)
-	  keys[keycc++] = '\n';
 	while (!feof(fp)
-	       && (cc = fread(keys + keycc, 1, keyalloc - keycc, fp)) > 0)
+	       && (cc = fread(keys + keycc, 1, keyalloc - 1 - keycc, fp)) > 0)
 	  {
 	    keycc += cc;
-	    if (keycc == keyalloc)
+	    if (keycc == keyalloc - 1)
 	      keys = xrealloc(keys, keyalloc *= 2);
 	  }
 	if (fp != stdin)
 	  fclose(fp);
-	/* Nuke the final newline to avoid matching a null string. */
-	if (keycc - oldcc > 0 && keys[keycc - 1] == '\n')
-	  --keycc;
-	keyfound = 1;
+	/* Append final newline if file ended in non-newline. */
+	if (oldcc != keycc && keys[keycc - 1] != '\n')
+	  keys[keycc++] = '\n';
 	break;
       case 'h':
 	no_filenames = 1;
@@ -897,7 +926,15 @@ warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"))
   if (show_help)
     usage(0);
 
-  if (!keyfound)
+  if (keys)
+    {
+      /* No keys were specified (e.g. -f /dev/null)/ Match nothing/ */
+      if (keycc == 0)
+        out_invert ^= 1;
+      else /* strip trailin newline. */
+        --keycc;
+    }
+  else
     if (optind < argc)
       {
 	keys = argv[optind++];
@@ -922,21 +959,24 @@ warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"))
   if (optind < argc)
     while (optind < argc)
       {
-	if (strcmp(argv[optind], "-"))
+	if (strcmp(argv[optind], "-") == 0)
 	  {
-	    /* On some platforms `open' will fail for a directory,
-	       so we stat the file before we attempt to open it.  */
+	    filename = _("(standard input)");
+	    desc = 0;
+	  }
+	else
+	  {
+#if defined(__MSDOS__) || defined(_WIN32)
 	    struct stat st;
-
-	    if (stat (argv[optind], &st) < 0 || S_ISDIR(st.st_mode))
+	    if (stat (argv[optind], &st) == 0 && S_ISDIR(st.st_mode))
 	      {
 		++optind;
 		continue;
 	      }
+#endif
+	    filename = argv[optind];
 	    desc = open(argv[optind], O_RDONLY);
 	  }
-	else
-	  desc = 0;
 	if (desc < 0)
 	  {
 	    if (!suppress_errors)
@@ -950,7 +990,6 @@ warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"))
 	    if (!isatty(desc))
 	      SET_BINARY(desc);
 #endif
-	    filename = desc == 0 ? _("(standard input)") : argv[optind];
 	    count = grep(desc);
 	    if (count_matches)
 	      {
@@ -966,9 +1005,9 @@ warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"))
 	      }
 	    else if (list_files == -1)
 	      printf("%s\n", filename);
+	    if (desc != 0)
+	      close(desc);
 	  }
-	if (desc != 0)
-	  close(desc);
 	++optind;
       }
   else
