@@ -1,5 +1,5 @@
 /* Portability cruft.  Include after config.h and sys/types.h.
-   Copyright (C) 1996 Free Software Foundation, Inc.
+   Copyright (C) 1996, 1998 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -17,12 +17,18 @@
    02111-1307, USA.  */
 
 #undef PARAMS
-#if __STDC__
-typedef void *ptr_t;
-#define PARAMS(x) x
+#if defined (__STDC__) && __STDC__
+# ifndef _PTR_T
+# define _PTR_T
+  typedef void * ptr_t;
+# endif
+# define PARAMS(x) x
 #else
-typedef char *ptr_t;
-#define PARAMS(x) ()
+# ifndef _PTR_T
+# define _PTR_T
+  typedef char * ptr_t;
+# endif
+# define PARAMS(x) ()
 #endif
 
 #define valloc system_valloc /* kludge */
